from pathlib import Path
import re


class Template:
    path: Path
    text: str

    def __init__(self, path: Path):
        self.path = path
        self._load()

    def _load(self):
        with open(self.path) as f:
            self.text = f.read()

    def replace(self, key, value) -> str:
        lookup = f"{{{{{key}}}}}"
        matches = re.findall(lookup, self.text)
        lookup = f"{{{{{key}=.*}}}}"  # match keys with defaults
        matches += re.findall(lookup, self.text)
        if not matches:
            raise ValueError(f"Could not find {key} key in template.")
        for match in matches:
            text = self.text.replace(match, value)
        return text

    def replace_defaults(self):
        lookup = f"{{.*=.*}}"
        matches = re.findall(lookup, self.text)
        for match in matches:
            key, value = match[2:-2].split("=")
            self.text = self.replace(key, value)

    def populate(self, defaults=True, **template_variables):
        for key, value in template_variables.items():
            self.text = self.replace(key, value)
        if not defaults:
            return
        self.replace_defaults()

    def write(self, outfile: Path):
        with open(outfile, "w") as f:
            f.write(self.text)



import json
import os
import glob
from pathlib import Path
from re import S
import textwrap
import typer
from typing import Optional, List, final
import shutil
import random
from dynamic_beast import create_dynamic_xml
from typing import List, Optional
from dataclasses import dataclass, field

from pybeast import Template, Command

app = typer.Typer()


def create_beast_run_command(
    dynamic_xml_path: Path,
    run_directory: Path,
    threads: int,
    json_path: Path,
    seed: int,
    resume: bool,
):
    """Create the BEAST run command"""

    cmd = Command("beast")

    if resume:
        cmd.add_arg("-resume")
    cmd.add_arg("-beagle")
    cmd.add_arg(f"-statefile {str(dynamic_xml_path).replace('.dynamic.', '.')}.state")
    cmd.add_arg(f"-seed {seed}")
    cmd.add_arg(f"-prefix {run_directory}/logs/")
    cmd.add_arg(f"-instances {threads}")
    cmd.add_arg(f"-threads {threads}")
    cmd.add_arg(f"-DF {json_path}")
    cmd.add_arg(f"-DFout {str(dynamic_xml_path).replace('.dynamic.', '.')}")
    cmd.add_arg(str(dynamic_xml_path))

    cmd.add_output_handler("2>&1 | tee", f"{run_directory}/{run_directory.stem}.out")

    return cmd


def populate_template(
    outfile: Path,
    cmd: Command,
    template_path: Path = None,
    template_variables: dict = None,
) -> Path:
    """Fill in a given template"""

    if template_path:
        template = Template(template_path)
        template.populate(BEAST=cmd.format(2), **template_variables)
        template.write(outfile)
    else:
        with open(outfile, "w") as f:
            f.write(cmd.format(2))

    return outfile


def create_dynamic_template(
    beast_xml_file: Path, outdir: Path, mc3: bool, ps: bool, ns: bool
) -> Path:
    """Given a BEAST2 xml file create a dynamic xml file and return it's path"""
    basename = beast_xml_file.stem
    dynamic_filename = Path(basename).with_suffix(".dynamic.xml")
    json_filename = Path(basename).with_suffix(".json")
    dynamic_outfile = f"{outdir}/{dynamic_filename}"
    json_outfile = f"{outdir}/{json_filename}"
    create_dynamic_xml(
        beast_xml_file,
        outfile=dynamic_outfile,
        json_out=json_outfile,
        mc3=mc3,
        ps=ps,
        ns=ns,
    )
    return Path(dynamic_outfile), Path(json_outfile)


def create_run_directory(
    beast_xml_path: Path,
    description: str,
    group: str,
    overwrite: bool,
    duplicate: int,
    resume: bool,
) -> Path:
    if resume:
        run_directory = Path(f"{beast_xml_path.parent}_RESUMED")
        run_directory_numbered = Path(f"{run_directory}_{duplicate:03d}")
    else:
        basename = beast_xml_path.stem
        run_directory = f"{basename}"
        if description:
            run_directory = f"{description}_{run_directory}"
        if group:
            run_directory = f"{group}/{run_directory}"
        run_directory_numbered = Path(f"{run_directory}_{duplicate:03d}")
    if overwrite and run_directory_numbered.exists():
        shutil.rmtree(run_directory_numbered)
    while run_directory_numbered.exists():
        duplicate += 1
        run_directory_numbered = Path(f"{run_directory}_{duplicate:03d}")
    os.makedirs(run_directory_numbered)
    os.makedirs(f"{run_directory_numbered}/logs")
    if resume:
        shutil.copytree(
            f"{beast_xml_path.parent}/logs/",
            f"{run_directory_numbered}/logs/",
            dirs_exist_ok=True,
        )
        state_file = glob.glob(f"{beast_xml_path.parent}/*.state")[0]
        shutil.copy(state_file, f"{run_directory_numbered}")
        shutil.copy(f"{beast_xml_path.parent}/seed.txt", f"{run_directory_numbered}")
    return Path(run_directory_numbered)


def set_dynamic_vars(json_path, samples, chain_length, dynamic_variables):
    with open(json_path) as f:
        data = json.load(f)
        if chain_length:
            data["mcmc.chainLength"] = chain_length
        else:
            chain_length = data["mcmc.chainLength"]
        if samples:
            sample_frequency = int(chain_length) // samples
            for key in data.keys():
                if key.startswith("treelog") and key.endswith("logEvery"):
                    data[key] = str(sample_frequency)
                if key.startswith("tracelog") and key.endswith("logEvery"):
                    data[key] = str(sample_frequency)
        data.update(dynamic_variables)
        json.dump(data, open(json_path, "w"), indent=4)


@app.command()
def main(
    beast_xml_path: Path,
    run: str = typer.Option(None, help="Run the run.sh file using this command."),
    resume: bool = typer.Option(False, help="Resume the specified run."),
    group: str = typer.Option(None, help="Group runs in this folder."),
    description: str = typer.Option("", help="Text to prepend to output folder name."),
    overwrite: bool = typer.Option(False, help="Overwrite run folder if exists."),
    seed: int = typer.Option(None, help="Seed to use in beast analysis."),
    duplicates: int = typer.Option(1, help="Number for duplicate runs to create."),
    dynamic_variable: Optional[List[str]] = typer.Option(
        None,
        "--dynamic-variable",
        "-d",
        help="Dynamic variable in the format <key>=<value>.",
    ),
    template: Path = typer.Option(
        None, help="Template for run.sh. Beast command is append to end of file."
    ),
    template_variable: Optional[List[str]] = typer.Option(
        None,
        "--template-variable",
        "-v",
        help="Template variable in the format <key>=<value>.",
    ),
    chain_length: int = typer.Option(None, help="Number of step in MCMC chain."),
    samples: int = typer.Option(None, help="Number of samples to collect."),
    threads: int = typer.Option(
        1,
        help="Number of threads and beagle instances to use (one beagle per thread).",
    ),
    mc3: bool = typer.Option(
        False, help="Use dynamic-beast to set default options for running MCMCMC."
    ),
    ps: bool = typer.Option(
        False, help="Use dynamic-beast to set default options for running PathSampler."
    ),
    ns: bool = typer.Option(
        False,
        help="Use dynamic-beast to set default options for running multi threaded nested sampling.",
    ),
):
    for i in range(duplicates):

        run_directory = create_run_directory(
            beast_xml_path,
            description,
            group=group,
            overwrite=overwrite,
            duplicate=i + 1,
            resume=resume,
        )

        dynamic_xml_path, json_path = create_dynamic_template(
            beast_xml_path, outdir=run_directory, mc3=mc3, ps=ps, ns=ns
        )
        dynamic_variables = {
            d.split("=")[0]: "".join(d.split("=")[1:]) for d in dynamic_variable
        }
        for dv in dynamic_variables:
            dynamic_variables[dv] = dynamic_variables[dv].replace(
                "{{run_directory}}", str(run_directory)
            )

        set_dynamic_vars(
            json_path,
            samples=samples,
            chain_length=chain_length,
            dynamic_variables=dynamic_variables,
        )

        beast_seed = str(random.randint(1, 10000000))
        if seed:
            beast_seed = str(seed)
        elif resume:
            with open(f"{run_directory}/seed.txt") as f:
                beast_seed = f.read()

        with open(f"{run_directory}/seed.txt", "w") as f:
            f.write(beast_seed)

        cmd_list = create_beast_run_command(
            dynamic_xml_path, run_directory, threads, json_path, beast_seed, resume
        )

        run_file = f"{run_directory}/run.sh"

        template_variables = {
            d.split("=")[0]: "".join(d.split("=")[1:]) for d in template_variable
        }

        populate_template(
            run_file,
            cmd_list,
            template_path=template,
            template_variables=template_variables,
        )
        typer.echo(f"Created run file -> {run_file}")

        if run:
            os.system(f"{run} {run_file}")



from dataclasses import dataclass, field
from typing import List, Optional


@dataclass
class Command:
    command: str
    final: Optional[str] = None
    args: List[str] = field(default_factory=list)

    def add_arg(self, arg: str):
        self.args.append(arg)

    def add_output_handler(self, pipe, value):
        self.final = f"{pipe} {value}"

    def format(self, indent=4):
        join_str = f' \\\n{" " * indent}'
        return f"""{self.command}{join_str}{join_str.join(self.args)} {self.final if self.final else ''}"""

    def __str__(self) -> str:
        return (
            f"{self.command} {' '.join(self.args)} {self.final if self.final else ''}"
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pybeast']

package_data = \
{'': ['*']}

install_requires = \
['dynamic-beast>=1.8.1', 'typer>=0.3.2']

entry_points = \
{'console_scripts': ['pybeast = pybeast.main:app']}

setup_kwargs = {
    'name': 'pybeast',
    'version': '0.3.2',
    'description': '',
    'long_description': '# pyBEAST\n\n[![PyPi](https://img.shields.io/pypi/v/pybeast.svg)](https://pypi.org/project/pybeast/)\n[![tests](https://github.com/Wytamma/pybeast/actions/workflows/test.yml/badge.svg)](https://github.com/Wytamma/pybeast/actions/workflows/test.yml)\n[![cov](https://codecov.io/gh/Wytamma/pybeast/branch/master/graph/badge.svg)](https://codecov.io/gh/Wytamma/pybeast)\n\nPyBEAST helps with running BEAST with best practices. Configure a beast run in a reproducible manner can be time consuming. pyBEAST is designed to making running beast as simple as possible. \n\n## Install\nInstall `pybeast` with pip (requires python -V >= 3.7).\n\n```bash\npip install pybeast\n```\n\n## Command line interface\n\n### Basic usage \n\n```bash\npybeast beast.xml\n```\n\n1. Create output folder and run command\n2. Ensures the run is self-contained and reproducible.\n\n```\npybeast --run bash beast.xml\n```\n\nThe --run flag tells pybeast how to run the run.sh file. \n\n### SLURM example \n\nThis example using the SLURM template in the examples folder to submit the beast run as a job.\n\n```bash\npybeast --run sbatch --template examples/slurm.template examples/beast.xml\n```\n\nAt a minimum the template must contain `{{BEAST}}` key. This will be replaced with the beast2 run command.\n\nHere we use the -v (--template-variable) option to request 4 cpus. \n\n```bash\npybeast --run sbatch --template examples/slurm.template -v cpus-per-task=4 exmaples/beast.xml\n```\n\nDefault template variables can be specified in the template in the format `{{<key>=<value>}}` e.g. {{cpus-per-task=4}}.\n\n## dynamic variables\n\nPyBEAST uses [dynamic-beast](https://github.com/Wytamma/dynamic-beast) to create dynamic xml files that can be modified at runtime. \n\nHere we use the -d (--dynamic-variable) option to set the chain length to 1000000. \n\n```bash\npybeast -d mcmc.chainLength=1000000 examples/beast.xml\n```\n\nThe dynamic variables are saved to a `.json` file in the run directory. This file can be further edited before runtime. At run time the values in the JSON file will be used in the analysis. \n\n## Example \n\n### BETS\n\nUse pybeast + feast to run BETS.\n\n```bash\nALIGNMENT=${1?Must provide an ALIGNMENT}\nfor XML_FILE in $(ls examples/BETS-templates/*)\ndo  \n    GROUP_NAME="BETS/$(basename "${ALIGNMENT}" .fasta)/$(basename "${XML_FILE}" .xml)"\n    pybeast \\\n        --run sbatch \\\n        --overwrite \\\n        --threads 8 \\\n        --duplicates 3 \\\n        --template examples/slurm.template \\\n        -v cpus-per-task=8 \\\n        --group $GROUP_NAME \\\n        -d "alignment=$ALIGNMENT" \\\n        -d "Date.delimiter=_" \\\n        -d "Date.dateFormat=yyyy/M/dd" \\\n        --ps \\\n        -d "ps.nrOfSteps=50" \\\n        -d "ps.chainLength=250000" \\\n        -d "ps.rootdir={{run_directory}}/logs" \\\n        $XML_FILE\ndone\n```\n\n## Help\n\n```\n❯ pybeast --help\nUsage: pybeast [OPTIONS] BEAST_XML_PATH\n\nArguments:\n  BEAST_XML_PATH  [required]\n\nOptions:\n  --run TEXT                    Run the run.sh file using this command.\n  --resume / --no-resume        Resume the specified run.  [default: no-\n                                resume]\n  --group TEXT                  Group runs in this folder.\n  --description TEXT            Text to prepend to output folder name.\n  --overwrite / --no-overwrite  Overwrite run folder if exists.  [default: no-\n                                overwrite]\n  --seed INTEGER                Seed to use in beast analysis.\n  --duplicates INTEGER          Number for duplicate runs to create.\n                                [default: 1]\n  -d, --dynamic-variable TEXT   Dynamic variable in the format <key>=<value>.\n  --template PATH               Template for run.sh. Beast command is append\n                                to end of file.\n  -v, --template-variable TEXT  Template variable in the format <key>=<value>.\n  --chain-length INTEGER        Number of step in MCMC chain.\n  --samples INTEGER             Number of samples to collect.\n  --threads INTEGER             Number of threads and beagle instances to use\n                                (one beagle per core). If not specified\n                                defaults to number of cores.  [default: 1]\n  --mc3 / --no-mc3              Use dynamic-beast to set default options for\n                                running MCMCMC.  [default: no-mc3]\n  --ps / --no-ps                Use dynamic-beast to set default options for\n                                running PathSampler.  [default: no-ps]\n  --ns / --no-ns                Use dynamic-beast to set default options for\n                                running multi threaded nested sampling.\n                                [default: no-ns]\n  --install-completion          Install completion for the current shell.\n  --show-completion             Show completion for the current shell, to copy\n                                it or customize the installation.\n  --help                        Show this message and exit.\n  ```',
    'author': 'Wytamma Wirth',
    'author_email': 'wytamma.wirth@me.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0',
}


setup(**setup_kwargs)
